#include "Avatar.h"
#include "Autonome.h"

#include "VueEspace.h"
#include "Deplacable.h"
#include "Nommable.h"
#include "TresorsSublimes.h"
//
#include "PartieEtudiante.h"
#include "ServicesSysteme.h"
#include "SingletonsOrdonnes.h"
#include "Singletons.h"
#include "Carte.h"
#include <iostream>
#include <sstream>
using std::ostringstream;
#include <algorithm>
using std::copy;
using std::count_if;
using std::fill;
#include <vector>
using std::vector;
#include <string>
using std::string;

//
// Fonction CreerMemoireAvatar()
// Fabrique une instance de MemoireAvatar
//
MemoireAvatar CreerMemoireAvatar
   (const point_t & pos, const Direction dir, const BoussoleMagique & boussole)
{
   MemoireAvatar mem;
   mem.direction_regard_ = dir;
   mem.position_ = pos;
   mem.boussole_magique_ = boussole;
   return mem;
}


class Avatar::Impl
   : public Autonome, public Deplacable, public Nommable
{
   friend class Avatar;
   DescripteurAvatar descr_;
   vector<point_t> pos_hurlements_;
   Mutex verrou_hurlements_;
   vector<MemoireAvatar> memoire_;
   enum { RUSH_ADRENALINE = 15 };
   int nb_iter_adrenaline_;
   DescripteurAvatar DeleguerControle
      (DescripteurAvatar desc, ostream &os) const
   {
      ControlerAvatar (os, nom (), descr_, desc, &memoire_[0],
                       static_cast<int>(memoire_.size ()));
      return desc;
   }
public:
   Impl (const string &nom, const point_t &pos)
      : Autonome (RythmeAvatarNormal ()), Deplacable (pos), Nommable (nom),
        nb_iter_adrenaline_ (0)
   {
      fill (descr_.etats_, descr_.etats_ + NB_ETATS_AVATAR, 0);
      descr_.direction_regard_ = Est;
   }
   void Agir ()
   {
      //
      // Prparation des tats
      //
      descr_.position_ = position ();
      RemplirZoneRegard (descr_);
      descr_.boussole_magique_ = ObtenirSingleton<
            TresorsSublimes
         >().DistancePlusProcheDe (descr_.position_);
      if (nb_iter_adrenaline_)
         --nb_iter_adrenaline_;
      descr_.nb_hurlements_proches_ = CompterHurlements ();
      if (descr_.nb_hurlements_proches_)
         nb_iter_adrenaline_ = RUSH_ADRENALINE;
      SetDelai (nb_iter_adrenaline_? RythmeAvatarRush () : RythmeAvatarNormal ());
      memoire_.push_back (
         CreerMemoireAvatar (
            descr_.position_, descr_.direction_regard_, descr_.boussole_magique_
         )
      );
      //
      // Dlgation  la partie tudiante
      //
      ostringstream sstr;
      DescripteurAvatar apres = DeleguerControle (descr_, sstr);
      //
      // Validation des dcisions
      //
      copy (apres.etats_, apres.etats_ + NB_ETATS_AVATAR, descr_.etats_);
      if (descr_.position_ != apres.position_)
         if (!deplacer (apres.position_))
            sstr << "Mouvement " << descr_.position_
                 << " --> " << apres.position_
                 << " refus pour l'avatar " << nom ();
         else
            descr_.position_ = apres.position_;
      descr_.direction_regard_ = apres.direction_regard_;
      EmettreMessage (sstr.str ());
   }
   void deplacerImpl (const point_t &Pos)
      { positionner (Pos); }
   void NoterHurlement (const point_t &Pos)
   {
      Autoverrou av (verrou_hurlements_);
      pos_hurlements_.push_back (Pos);
   }
   int CompterHurlements ()
   {
      class AssezProche
      {
         point_t pos_;
      public:
         AssezProche (const point_t &pos)
            : pos_(pos)
         {
         }
         bool operator() (const point_t &pos) const
            { return pos - pos_ < RAYON_OUIE_AVATAR; }
      };
      Autoverrou av (verrou_hurlements_);
      int Nb = static_cast<int>(count_if
         (pos_hurlements_.begin (), pos_hurlements_.end (), AssezProche (position ())));
      pos_hurlements_.clear ();
      return Nb;
   }
};

Avatar::Impl *Avatar::Creer (const string &nom, const point_t &pos, Avatar *pAbonne)
{
   Avatar::Impl *p = new Avatar::Impl (nom, pos);
   p->Abonner (static_cast<AbonneDeplacable *>(pAbonne));
   p->Demarrer ();
   return p;
}

void Avatar::Liberer (const Avatar::Impl *p) throw()
   { delete p; }

Avatar::~Avatar () throw()
   { pimpl_->Desabonner (this); }

void Avatar::deplacerImpl (const point_t &pos)
{
   pimpl_->deplacerImpl (pos);
   positionner (pos);
}

const string & Avatar::nom () const throw ()
   { return pimpl_->nom (); }

void Avatar::NoterHurlement (const point_t &pos)
   { pimpl_-> NoterHurlement (pos); }

template <>
   Deplacable * Avatar::to_<Deplacable>() const throw ()
      { return pimpl_.GetPartageable().get(); }
template <>
   Positionnable* Avatar::to_<Positionnable>() const throw ()
      { return pimpl_.GetPartageable().get(); }