#include "Carte.h"
#include <algorithm>
using std::for_each;
using std::fill;
#include <utility>
using std::max;
#include <vector>
using std::vector;
#include <string>
using std::string;

//
// Piarkkk!
//
void CalculerObstructions (ZoneRegard &cr)
{
   // +---+---+---+---+---+
   // |   |   |   |   |   |
   // +---+---+---+---+---+
   // |   |   |   |   |   |
   // +---+---+---+---+---+
   // |   |   | A |   |   |
   // +---+---+---+---+---+
   // |   |   |   |   |   |
   // +---+---+---+---+---+
   // |   |   |   |   |   |
   // +---+---+---+---+---+
   //
   // Coin haut droit
   if (EstSymboleMur (cr.zone_[1][3]))
      cr.zone_[0][4] = cr.zone_[0][3] = cr.zone_[1][4] = SYMBOLE_INCONNU;
   //
   // Coin haut gauche
   if (EstSymboleMur (cr.zone_[1][1]))
      cr.zone_[0][0] = cr.zone_[0][1] = cr.zone_[1][0] = SYMBOLE_INCONNU;
   //
   // Coin bas gauche
   if (EstSymboleMur (cr.zone_[3][1]))
      cr.zone_[4][0] = cr.zone_[4][1] = cr.zone_[3][0] = SYMBOLE_INCONNU;
   //
   // Coin bas droit
   if (EstSymboleMur (cr.zone_[3][3]))
      cr.zone_[4][4] = cr.zone_[4][3] = cr.zone_[3][4] = SYMBOLE_INCONNU;
   //
   // Est
   if (EstSymboleMur (cr.zone_[2][3]))
      cr.zone_[1][4] = cr.zone_[2][4] = cr.zone_[3][4] = SYMBOLE_INCONNU;
   //
   // Nord
   if (EstSymboleMur (cr.zone_[1][2]))
      cr.zone_[0][2] = cr.zone_[0][3] = cr.zone_[0][4] = SYMBOLE_INCONNU;
   //
   // Ouest
   if (EstSymboleMur (cr.zone_[2][1]))
      cr.zone_[1][0] = cr.zone_[2][0] = cr.zone_[3][0] = SYMBOLE_INCONNU;
   //
   // Sud
   if (EstSymboleMur (cr.zone_[4][2]))
      cr.zone_[4][2] = cr.zone_[4][3] = cr.zone_[4][4] = SYMBOLE_INCONNU;
}

struct NoOp
{
   template <class T>
      T operator () (T val)
         { return val; }
};

template <int LAR, int HAU>
   struct RemplirZoneImpl
   {
      typedef position_traits::coord_t coord_t;
      typedef position_traits::point_t point_t;
      static const coord_t DX = LAR / 2;
      static const coord_t DY = HAU / 2;
      template <class Z, class TransOp, class CType, class Jeton>
         void operator()
            (const point_t &pos, Z &zone, Carte<CType> &carte, TransOp oper, Jeton jeton)
         {
            const int BASE_X = pos.X - DX;
            const int BASE_Y = pos.Y - DY;
            for (int i = 0; i < HAU; ++i)
               for (int j = 0; j < LAR; ++j)
                  zone.zone_[i][j] = oper
                     (carte.symbole(point_t(BASE_X + j, BASE_Y + i), jeton));
         }
   };
 
template <int LAR, int HAU, class Z, class TransOp>
   class RemplirZoneSynchro
   {
      const point_t &pos_;
      Z &zone_;
      TransOp oper_;
   public:
      RemplirZoneSynchro (const point_t &pos, Z &zone, TransOp oper)
         : pos_(pos), zone_(zone), oper_(oper)
      {
      }
      template <class CType, class Jeton>
         void operator () (Carte<CType>& carte, Jeton jeton)
         {
            RemplirZoneImpl<LAR, HAU> rz;
            rz (pos_, zone_, carte, oper_, jeton);
         }
   };

template <int LAR, int HAU, class CType, class Z, class TransOp>
   void RemplirZone (const point_t &pos, Z &zone, TransOp oper)
   {
      Carte<CType>::Get()->injecter
         (RemplirZoneSynchro<LAR, HAU, Z, TransOp> (pos, zone, oper));
   }


void RemplirZoneRegard (DescripteurAvatar &desc)
{
   RemplirZone<LARGEUR_ZONE_REGARD, HAUTEUR_ZONE_REGARD, char>
      (desc.position_, desc.observation_, NoOp());
}

void RemplirZoneOdorat (DescripteurMonstre &desc)
{
   RemplirZone<LARGEUR_ZONE_ODORAT, HAUTEUR_ZONE_ODORAT, char>
      (desc.position_, desc.chair_fraiche_, EstSymboleAvatar);
   RemplirZone<LARGEUR_ZONE_ODORAT, HAUTEUR_ZONE_ODORAT, char>
      (desc.position_, desc.chair_moite_, EstSymboleMonstre);
}

#include "ServicesSysteme.h"
#include "Avatar.h"
#include "TroupeHeroique.h"
#include "SingletonsOrdonnes.h"
#include "Singletons.h"

void Hurler (const string &nom, const DescripteurMonstre &desc)
{
   typedef position_traits::point_t point_t;
   EmettreMessage ("Le monstre " + nom + " vient de hurler");
   class EntendreHurlement
   {
      point_t pos_;
   public:
      EntendreHurlement (const point_t &pos)
         : pos_(pos)
      {
      }
      void operator () (Avatar *p)
         { p->NoterHurlement (pos_); }
   };
   ObtenirSingleton<TroupeHeroique>().Propager (EntendreHurlement (desc.position_));
}
