#ifndef CARTE_H
#define CARTE_H

#include "Incopiable.h"
#include "PositionTraits.h"
#include "PartieEtudiante.h"
#include "PtrPartage.h"
#include "Dimension.h"
#include "Mutex.h"
#include <vector>
using std::vector;
#include <string>
using std::basic_string;
#include <algorithm>
using std::for_each;

template <class CType>
   class PlusLongueChaine
   {
   public:
      typedef typename basic_string<CType> str_type;
      typedef typename str_type::size_type size_type;
   private:
      size_type lg_;
   public:
      PlusLongueChaine () throw()
         : lg_(0)
      {
      }
      void operator () (const basic_string<CType> &s) throw()
         { lg_ = max (lg_, s.length ()); }
      size_type valeur () const throw()
         { return lg_; }
   };

//
// ICI: faire en sorte que les mthodes synchronises soient qualifies
// de volatile et qu'il existe des dclinaisons non volatiles des mmes
// mthodes pour une ventuelle version monoprogramme du systme
//
template <class CType>
   class Carte
      : Incopiable
   {
   public:
      typedef basic_string<CType> str_type;
      class CarteHolder
      {
         PtrPartage<Carte> carte_;
      public:
         CarteHolder ()
            : carte_ (new Carte)
         {
         }
         PtrPartage<Carte> GetCarte ()
            { return carte_; }
      };
      vector<str_type> cases_;
   public:
      typedef position_traits::coord_t coord_t;
      typedef position_traits::point_t point_t;
   private:
      Mutex acces_carte_;
      Dimension<coord_t> dim_;
      Carte ()
         : dim_(coord_t(), coord_t())
      {
      }
      friend class Donjon;
      void Attribuer (const vector<str_type> &c)
      {
         cases_ = c;
         dim_ = Dimension<coord_t>
            (for_each (c.begin (), c.end (), PlusLongueChaine<CType> ()).valeur (),
             c.size ());
         class HarmoniserLargeur
         {
            coord_t largeur_;
         public:
            HarmoniserLargeur (coord_t lar)
               : largeur_(lar)
            {
            }
            void operator () (string &s)
            {
               while (s.length () < static_cast<str_type::size_type> (largeur_))
                  s.push_back (SYMBOLE_VIDE);
            }
         };
         for_each (cases_.begin (), cases_.end (),
                   HarmoniserLargeur (dimension().largeur ()));
      }
      class PreuveDeSynchro {};
   public:
      template <class Op>
         Op injecter (Op oper)
         {
            Autoverrou av (acces_carte_); // ICI: rflchir au niveau de granularit requis
            oper (*this, PreuveDeSynchro ());
            return oper;
         }
      static PtrPartage<Carte> Get ()
      {
         static CarteHolder Singleton;
         return Singleton.GetCarte ();
      }
      void echanger (const point_t &p0, const point_t &p1, PreuveDeSynchro) throw()
      {
         std::swap (symboleRef(p0), symboleRef(p1));
      }
      void remplacer (const point_t &p, const CType sym, PreuveDeSynchro) throw()
      {
         symboleRef(p) = sym;
      }
      const vector<str_type> & Cases (PreuveDeSynchro) const throw()
         { return cases_; }
      CType Case (const point_t &pt, PreuveDeSynchro) const throw()
         { return cases_[pt.Y][pt.X]; }
      const Dimension<coord_t> dimension () const throw ()
         { return dim_; }
      bool EstDans (const point_t &pos, PreuveDeSynchro) const throw()
      {
         return pos.X >= 0 && pos.X < dimension().largeur () &&
                pos.Y >= 0 && pos.Y < dimension().hauteur ();
      }
      bool EstDans (const point_t &pos) const throw()// ICI: faire disparatre
      {
         return pos.X >= 0 && pos.X < dimension().largeur () &&
                pos.Y >= 0 && pos.Y < dimension().hauteur ();
      }
      point_t centre () const
         { return point_t (dimension().hauteur () / 2, dimension().largeur () / 2); }
      CType symbole (const point_t &pos, PreuveDeSynchro) const throw()
         { return EstDans (pos)? cases_[pos.Y][pos.X] : SYMBOLE_INCONNU; }
      //CType symbole (const point_t &pos) const throw() // ICI: faire disparatre
      //   { return EstDans (pos)? cases_[pos.Y][pos.X] : SYMBOLE_INCONNU; }
   private:
      CType & symboleRef (const point_t &pos) throw()
         { return cases_[pos.Y][pos.X]; }
   };

void RemplirZoneRegard (DescripteurAvatar &desc);
void RemplirZoneOdorat (DescripteurMonstre &desc);

#endif
