#ifndef CATALOGUE_H
#define CATALOGUE_H

#include <vector>
using std::vector;
#include <string>
using std::basic_string;
#include <algorithm>
using std::find_if;
#include <iosfwd>
using std::istream;

template <class CType>
   class Catalogue
   {
   public:
      struct Element
      {
         CType symbole_;
         typedef basic_string<CType> str_type;
         str_type nom_;
         Element (const CType symbole = CType (), const str_type &nom = str_type())
            : symbole_ (symbole), nom_(nom)
         {
         }
         friend istream & operator>> (istream &is, Element &e)
         {
            istream::sentry verrou (is);
            if (verrou)
            {
               CType symbole;
               str_type nom;
               is >> symbole >> nom;
               e = Element (symbole, nom);
            }
            return is;
         }
         class MemeSymbole
         {
            CType symbole_;
         public:
            MemeSymbole (const CType c)
               : symbole_ (c)
            {
            }
            bool operator () (const Element &e) const throw ()
               { return e.symbole_ == symbole_; }
         };
      };
      typedef typename
         vector<Element>::size_type size_type;
      typedef typename
         vector<Element>::iterator iterator;
      typedef typename
         vector<Element>::const_iterator const_iterator;
      iterator begin () throw ()
         { return catalogue_.begin (); }
      const_iterator begin () const throw ()
         { return catalogue_.begin (); }
      iterator end () throw ()
         { return catalogue_.end (); }
      const_iterator end () const throw ()
         { return catalogue_.end (); }
      void push_back (const Element &e)
         { catalogue_.push_back (e); }
      bool empty () const throw ()
         { return catalogue_.empty (); }
      size_type size () const throw ()
         { return catalogue_.size (); }
      iterator trouver (const char Symbole) throw ()
         { return find_if (begin (), end (), Element::MemeSymbole (Symbole)); }
      const_iterator trouver (const char Symbole) const throw ()
         { return find_if (begin (), end (), Element::MemeSymbole (Symbole)); }
      bool possede (const char Symbole) const throw ()
         { return trouver (Symbole) != end (); }
   private:
      vector<Element> catalogue_;
   };

template <class CType>
   typename Catalogue<CType>::Element::MemeSymbole
      CreerTrouverMemeSymbole (const CType c)
         { return typename Catalogue<CType>::Element::MemeSymbole (c); }

template <class CType>
   bool TrouverDansCatalogue (const Catalogue<CType> &cat, const CType elem)
      { return cat.trouver (elem) != cat.end (); }


#endif