#include "Donjon.h"
#include "HordeMalefique.h"
#include "Monstre.h"
#include "TroupeHeroique.h"
#include "Avatar.h"
#include "TresorsSublimes.h"
#include "Tresor.h"
#include "SingletonsOrdonnes.h"
#include "Singletons.h"
#include "locale_helper.h"
#include <fstream>
using std::basic_ifstream;
#include <sstream>
using std::basic_stringstream;
#include <vector>
using std::vector;
#include <string>
using std::string;
using std::getline;
#include <utility>
using std::swap;
#include <algorithm>
using std::find;
using std::find_if;
using std::for_each;

#include <string>
using std::string;
#include <locale>
using std::locale;

const Donjon::str_type Donjon::CARTE_DEFAUT =
   convertir<Donjon::CType>(string ("Carte.txt"));

#pragma warning (disable : 4290)

#include "ServicesSysteme.h"

Donjon::Donjon (const str_type & nomCarte)
   throw (CarteInexistante, CarteVide, CarteInvalide)
{
   basic_ifstream<CType> ifs (nomCarte.c_str ());
   if (!ifs)
   {
      EmettreMessage ("Carte vide");
      ProvoquerFinDePartie (Partie::ERREUR_INITIALISATION);
      throw CarteInexistante ();
   }
   vector<str_type> v;
   for (str_type ligne; getline (ifs, ligne); v.push_back (ligne))
      ;
   if (v.empty ())
   {
      EmettreMessage ("Carte vide");
      ProvoquerFinDePartie (Partie::ERREUR_INITIALISATION);
      throw CarteVide ();
   }
   ConsommerCarte (v);
   if (!NbHerosValide ())
   {
      EmettreMessage ("Il faut un et un seul hros dans cette aventure");
      ProvoquerFinDePartie (Partie::ERREUR_INITIALISATION);
      throw CarteInvalide ();
   }
}

Donjon::str_type Donjon::ConsommerLigneCarte
   (const str_type &ligne, const vector<str_type>::size_type noLigne)
{
   typedef Positionnable::point_t point_t;
   typedef point_t::coord_t coord_t;
   str_type s = ligne;
   for (str_type::size_type i = 0; i < s.size (); ++i)
   {
      const point_t pos
         (static_cast<coord_t>(i), static_cast<coord_t>(noLigne));
      const CType SYMBOLE = s[i];
      // ICI: je prsume l'unicit du symbole SYMBOLE
      if (ObtenirCatalogue<Avatar>().possede (SYMBOLE))
      {
         Avatar *p = new Avatar (ObtenirCatalogue<Avatar>().trouver (SYMBOLE)->nom_, pos);
         ObtenirSingleton<TroupeHeroique>().Ajouter (p);
         deplacables_.push_back (p->to_<Deplacable>());
      }
      else if (ObtenirSingleton<Catalogues>().catalogue<Monstre> ().possede (SYMBOLE))
      {
         Catalogues::PatchMonstres patch =
            ObtenirSingleton<Catalogues>().TrouverMonstre (SYMBOLE);
         Monstre *p = new Monstre (patch.nom_, pos, patch.mode_);
         ObtenirSingleton<HordeMalefique>().Ajouter (p);
         deplacables_.push_back (p->to_<Deplacable>());
      }
      else if (ObtenirCatalogue<Tresor>().possede (SYMBOLE))
      {
         Tresor *p = new Tresor (ObtenirCatalogue<Tresor> ().trouver (SYMBOLE)->nom_, pos);
         ObtenirSingleton<TresorsSublimes>().Ajouter (p);
      }
   }
   return s;
}

bool Donjon::ConsommerLigneLegende (const str_type &ligne)
   throw (CarteInvalide)
{
   typedef Catalogues::catalog_type catalog_type;
   const locale loc ("");
   static const str_type
      ETIQUETTE_AVATAR  = convertir<CType>(string ("Avatar"), loc),
      ETIQUETTE_CARTE   = convertir<CType>(string ("Carte"), loc),
      ETIQUETTE_MONSTRE = convertir<CType>(string ("Monstre"), loc),
      ETIQUETTE_TRESOR  = convertir<CType>(string ("Tresor"), loc);
   basic_stringstream<CType> sstr;
   sstr << ligne;
   str_type etiquette;
   sstr >> etiquette;
   if (etiquette == ETIQUETTE_CARTE)
      return false;
   catalog_type::Element e;
   sstr >> e;
   if (etiquette == ETIQUETTE_AVATAR)
      ObtenirSingleton<Catalogues>().Ajouter<Avatar> (e);
   else if (etiquette == ETIQUETTE_TRESOR)
      ObtenirSingleton<Catalogues>().Ajouter<Tresor> (e);
   else if (etiquette == ETIQUETTE_MONSTRE)
   {
      Monstre::mode_deplacement_t mode = Monstre::mode_deplacement_t();
      if (sstr) sstr >> mode;
      ObtenirSingleton<Catalogues>().Ajouter<Monstre> (e, mode);
   }
   else
   {
      EmettreMessage
         ("La carte suggre est invalide: tiquette " + etiquette + " inconnue");
      ProvoquerFinDePartie (Partie::ERREUR_INITIALISATION);
      throw CarteInvalide ();
   }
   return true;
}

void Donjon::ConsommerCarte (const vector<str_type> &v)
   throw (CarteInvalide, CarteVide)
{
   bool ConsommationLegende = true;
   vector<str_type> carte;
   for (vector<str_type>::const_iterator itt = v.begin (); itt != v.end (); ++itt)
   {
      if (!ConsommationLegende)
         carte.push_back (ConsommerLigneCarte (*itt, carte.size ()));
      else if (!itt->empty ())
         ConsommationLegende = ConsommerLigneLegende (*itt);
   }
   if (carte.empty ())
   {
      EmettreMessage ("La carte suggre est vide");
      ProvoquerFinDePartie (Partie::ERREUR_INITIALISATION);
      throw CarteVide ();
   }
   Carte<CType>::Get ()->Attribuer (carte);
   // abonner chaque avatar
   class AutoAbonner
   {
      Donjon *abonne_;
   public:
      AutoAbonner (Donjon *p)
         : abonne_ (p)
      {
      }
      void operator() (Deplacable *p)
      {
         p->Abonner (static_cast<AutoriteDeplacement *>(abonne_));
         p->Abonner (static_cast<AbonneDeplacable *>(abonne_));
      }
   };
   for_each (deplacables_.begin (), deplacables_.end (), AutoAbonner (this));
}

#include <iterator>
using std::distance;
#include <iostream>
using std::basic_ostream;


class ProjectionSynchronisee
{
public:
   typedef Donjon::CType CType;
private:
   basic_ostream<CType> &os_;
   const Dimension<coord_t> &dim_;
public:
   ProjectionSynchronisee
      (basic_ostream<CType> &os, const Dimension<coord_t> &dim)
         : os_(os), dim_(dim)
      {
      }
   template <class JetonPrivilege>
      void operator () (Carte<char> &carte, JetonPrivilege jeton)
      {
         class ContientAvatar
         {
            const Catalogues::catalog_type &cat_;
         public:
            ContientAvatar (const Catalogues::catalog_type &cat)
               : cat_(cat)
            {
            }
            bool operator () (const string &s) const
            {
               return cat_.empty ()?
                  false : find_if (s.begin (), s.end (), EstSymboleAvatar) != s.end ();
            }
         };
         typedef basic_string<CType> str_type;
         const vector<str_type> &cases = carte.Cases(jeton);
         vector<str_type>::const_iterator
            debut = cases.begin ();
         vector<str_type>::const_iterator itt =
            find_if (debut, cases.end (), ContientAvatar (ObtenirCatalogue<Avatar> ()));

         point_t centre = (itt != cases.end ())?
            point_t (itt->find (ObtenirCatalogue<Avatar> ().begin ()->symbole_),
                                distance (debut, itt)) :
            point_t (carte.centre ());

         const point_t BASE
            (centre.X - dim_.largeur() / 2, centre.Y - dim_.hauteur() / 2);
         for (coord_t j = 0; j < dim_.hauteur(); ++j)
         {
            for (coord_t i = 0; i < dim_.largeur(); ++i)
            {
               const point_t OFFS (i + BASE.X, j + BASE.Y);
               // hors de la carte?
               if (!carte.EstDans (point_t (OFFS.X, OFFS.Y)) ||
                   static_cast<coord_t>(cases.size()) <= OFFS.Y ||
                   static_cast<coord_t>(cases[OFFS.Y].length ()) <= OFFS.X)
                  os_ << SYMBOLE_VIDE;
               else
                  os_ << carte.Case(OFFS, jeton);
            }
            os_ << '\n';
         }
      }
   basic_ostream<CType> & resultat () throw ()
      { return os_; }
};

basic_ostream<Donjon::CType> &Donjon::projeter
   (basic_ostream<CType> &os, const Dimension<coord_t> &dim) const
{
   return Carte<char>::Get ()->injecter (ProjectionSynchronisee (os, dim)).resultat ();
}


#include <exception>
using std::exception;

class FinSoudaine
   : public exception
{
public:
   typedef Partie::resultat_t resultat_t;
private:
   string texte_; // hum...
   resultat_t code_;
public:
   FinSoudaine (const string &texte, const resultat_t code)
      : texte_(texte), code_(code)
   {
   }
   const char * what () const throw()
      { return texte_.c_str (); }
   resultat_t raison () const throw()
      { return code_; }
};

class MouvementHerosSynchronise
{
   const string &nom_;
   const point_t &avant_, &apres_;
public:
   MouvementHerosSynchronise
      (const string &nom, const point_t & avant, const point_t & apres)
         : nom_(nom), avant_(avant), apres_(apres)
      {
      }
   template <class JetonPrivilege>
      void operator () (Carte<char> &carte, JetonPrivilege jeton)
      {
         Monstre *pDanger = ObtenirSingleton<HordeMalefique>().Identifier (apres_);
         if (pDanger)
            throw FinSoudaine("L'avatar " + convertir<char>(nom_, locale("")) +
                              " a t dvor par " + convertir<char>(pDanger->nom (), locale("")) + "!",
                              Partie::AVATAR_DEVORE);
         if (EstSymboleTresor (carte.symbole (apres_, jeton)))
         {
            carte.echanger  (avant_, apres_, jeton);
            carte.remplacer (avant_, SYMBOLE_VIDE, jeton); // trsor mang!
            throw FinSoudaine("L'avatar " + convertir<char>(nom_, locale("")) +
                              " a pris le trsor!",
                              Partie::TRESOR_TROUVE);
         }
         carte.echanger (avant_, apres_, jeton);
      }
};


void Donjon::MouvementHeros
   (const str_type &nom, const point_t & avant, const point_t & apres)
{
   Carte<char>::Get ()->injecter (MouvementHerosSynchronise (nom, avant, apres));
}


class MouvementMonstreSynchronise
{
   const string &nom_;
   const point_t &avant_, &apres_;
public:
   MouvementMonstreSynchronise
      (const string &nom, const point_t & avant, const point_t & apres)
         : nom_(nom), avant_(avant), apres_(apres)
      {
      }
   template <class JetonPrivilege>
      void operator () (Carte<char> &carte, JetonPrivilege jeton)
      {
         Avatar *pCible = ObtenirSingleton<TroupeHeroique>().Identifier (apres_);
         if (pCible)
            throw FinSoudaine ("L'avatar " + convertir<char>(pCible->nom (), locale("")) +
                               " a t dvor par " + nom_ + "!",
                               Partie::AVATAR_DEVORE);
         if (!EstSymboleVide (carte.symbole (apres_, jeton)))
         {
            Tresor *pTresor = ObtenirSingleton<TresorsSublimes>().Identifier (apres_);
            EmettreMessage ("Le monstre " + convertir<char>(nom_, locale("")) +
                            " vient de dplacer le trsor " +
                            pTresor->nom () + "!");
            pTresor->positionner (avant_);
         }
         carte.echanger (avant_, apres_, jeton);
      }
};

void Donjon::MouvementMonstre
   (const str_type &nom, const point_t & avant, const point_t & apres)
{
   Carte<char>::Get ()->injecter (MouvementMonstreSynchronise (nom, avant, apres));
}

bool Donjon::TrouverActeur (const Positionnable *p) const throw ()
   { return find (deplacables_.begin (), deplacables_.end (), p) != deplacables_.end (); }


void Donjon::Mouvement (const Positionnable *p, const point_t avant)
{
   try
   {
      SectionCritique::Pogneur pogn (critsect_);
      if (!TrouverActeur (p))
         throw FinSoudaine ("Erreur grave, acteur invalide", Partie::ERREUR_EXECUTION);
      const point_t POS = p->position ();
      Monstre *pMonstreActeur = ObtenirSingleton<HordeMalefique>().Identifier (p);
      Avatar *pAvatarActeur = ObtenirSingleton<TroupeHeroique>().Identifier (p);
      if (pAvatarActeur)
         MouvementHeros (pAvatarActeur->nom (), avant, POS);
      else if (pMonstreActeur)
         MouvementMonstre (pMonstreActeur->nom (), avant, POS);
      else // mur? trsor?
         throw FinSoudaine ("Erreur grave: un objet immobile semble avoir tent de bouger!", Partie::ERREUR_EXECUTION);
   }
   catch (const FinSoudaine & fin)
   {
      EmettreMessage (fin.what ());
      ProvoquerFinDePartie (fin.raison ());
   }
}


class AutoriserMouvementSynchronise
{
   const Positionnable *p_;
   const point_t &apres_;
   bool resultat_;
public:
   AutoriserMouvementSynchronise
      (const Positionnable *p, const point_t &apres)
         : p_(p), apres_(apres), resultat_(false)
      {
      }
   template <class JetonPrivilege>
      bool EstInfranchissable (Carte<char> &carte, const point_t &pos, JetonPrivilege jeton)
      {
         const symbole_t SYM = carte.symbole (pos, jeton);
         return (!carte.EstDans (pos) ||
                 !(EstSymboleVide (SYM) || EstSymboleAvatar (SYM) ||
                   EstSymboleMonstre (SYM) || EstSymboleTresor (SYM)));
      }
   template <class JetonPrivilege>
      void operator () (Carte<char> &carte, JetonPrivilege jeton)
      {
         resultat_=
            !EstInfranchissable (carte, apres_, jeton) &&
            !(ObtenirSingleton<HordeMalefique>().Identifier (p_) &&
              EstSymboleMonstre (carte.symbole (apres_, jeton))) &&
            (apres_ - p_->position () == 1);
      }
   bool resultat () const throw()
      { return resultat_; }
};


bool Donjon::AutoriserMouvement (const Positionnable *p, const point_t &apres)
{
   return Carte<char>::Get ()->injecter (AutoriserMouvementSynchronise (p, apres)).resultat ();
}

#pragma warning (default : 4290)

Donjon::~Donjon () throw()
{
   class AutoDesabonner
   {
      Donjon *abonne_;
   public:
      AutoDesabonner (Donjon *p) throw()
         : abonne_ (p)
      {
      }
      void operator() (Deplacable *p)
      {
         p->Desabonner (static_cast<AutoriteDeplacement *>(abonne_));
         p->Desabonner (static_cast<AbonneDeplacable *>(abonne_));
      }
   };
   for_each (deplacables_.begin (), deplacables_.end (), AutoDesabonner (this));
}

