#ifndef EVENEMENT_H
#define EVENEMENT_H

class IEvenement
{
public:
   class ReinitialisationAutomatique {};
   class ReinitialisationManuelle  {};
   virtual bool Attendre () const volatile = 0;
   virtual bool Attendre (const int Ms) const volatile = 0;
   virtual void Provoquer () const volatile = 0;
   virtual void Reinitialiser () const volatile = 0;
   static const volatile IEvenement *Creer (ReinitialisationAutomatique);
   static const volatile IEvenement *Creer (ReinitialisationManuelle);
   static void Liberer (const volatile IEvenement *) throw();
protected:
   virtual ~IEvenement () throw () = 0;
};

class Evenement
{
   const volatile IEvenement *evenement_;
public:
   typedef IEvenement::ReinitialisationAutomatique ReinitialisationAutomatique;
   typedef IEvenement::ReinitialisationManuelle ReinitialisationManuelle;
   template <class Categorie>
      Evenement (Categorie categ)
          : evenement_ (IEvenement::Creer(categ))
      {
      }
   ~Evenement () throw()
      { IEvenement::Liberer(evenement_); }
   bool Attendre () const volatile
      { return evenement_->Attendre (); }
   bool Attendre (const int Ms) const volatile
      { return evenement_->Attendre (Ms); }
   void Provoquer () const volatile
      { return evenement_->Provoquer (); }
   void Reinitialiser () const volatile
      { return evenement_->Reinitialiser (); }
};

#endif