#ifndef GROUPEMENT_ACTIF_H
#define GROUPEMENT_ACTIF_H

#include "Incopiable.h"
#include "PositionTraits.h"
#include "Mutex.h"
#include "Carte.h"
#include "PtrPartage.h"
#include "SingletonWrapper.h"
#include <algorithm>
using std::for_each;
using std::find_if;
#include <vector>
using std::vector;

class Positionnable;

template <class T, class CType>
   class GroupementActif
      : Incopiable
   {
      Mutex verrou_;
      vector<T *> groupe_;
      GroupementActif ()
         : lien_carte_ (Carte<CType>::Get ())
      {
      }
      friend struct SingletonWrapper<GroupementActif>;
      PtrPartage <Carte<CType> > lien_carte_;
   public:
      ~GroupementActif () throw()
      {
         Autoverrou av (verrou_);
         for_each (groupe_.begin (), groupe_.end (), Detruire ());
      }
      void Ajouter (T *p)
      {
         if (p)
         {
            Autoverrou av (verrou_);
            groupe_.push_back (p);
         }
      }
      T* Identifier (const Positionnable *p)
      {
         class Trouveur
         {
            const Positionnable *p_;
         public:
            Trouveur(const Positionnable *p)
               : p_(p)
            {
            }
            bool operator () (T *candidat)
               { return candidat->to_<Positionnable>() == p_; }
         };
         vector<T*>::iterator itt =
            find_if (groupe_.begin (), groupe_.end (), Trouveur(p));
         return itt != groupe_.end ()? *itt : 0;
      }
      T * Identifier (const position_traits::point_t &pos)
      {
         typedef position_traits::point_t point_t;
         class Trouveur
         {
            const point_t pos_;
         public:
            Trouveur(const point_t &pos)
               : pos_(pos)
            {
            }
            bool operator () (T* candidat)
               { return candidat->to_<Deplacable> ()->position () == pos_; }
         };
         vector<T*>::iterator itt =
            find_if (groupe_.begin (), groupe_.end (), Trouveur(pos));
         return itt != groupe_.end ()? *itt : 0;
      }
      template <class Op>
         void Propager (Op oper)
         {
            Autoverrou av (verrou_);
            for_each (groupe_.begin (), groupe_.end (), oper);
         }
   };

#endif
