#include "Monstre.h"
#include "Autonome.h"

#include "VueEspace.h"
#include "Deplacable.h"
#include "Carte.h"
#include "Nommable.h"
//
#include "ServicesSysteme.h"
#include "PartieEtudiante.h"
#include <sstream>
using std::stringstream;
#include <deque>
using std::deque;
#include <iostream>
#include <algorithm>
using std::fill;
using std::copy;
#include <string>
using std::string;

class Monstre::Impl
   : public Autonome, public Deplacable, public Nommable
{
public:
   typedef Monstre::mode_deplacement_t mode_deplacement_t;
private:
   string nom_;
   DescripteurMonstre descr_;
   friend class Monstre;
   deque<MemoireMonstre> memoire_;
   mode_deplacement_t mode_;
public:
   Impl (const string &nom, const point_t &pos, const mode_deplacement_t mode)
      : Deplacable (pos), Autonome (RythmeMonstreNormal ()),
        Nommable (nom), mode_ (mode)
   {
      fill (descr_.etats_, descr_.etats_ + NB_ETATS_MONSTRE, 0);
      descr_.direction_regard_ = Nord;
      for (int i = 0; i < HAUTEUR_ZONE_ODORAT; ++i)
      {
         fill (descr_.chair_fraiche_.zone_[i],
               descr_.chair_fraiche_.zone_[i] + LARGEUR_ZONE_ODORAT, false);
         fill (descr_.chair_moite_.zone_[i],
               descr_.chair_moite_.zone_[i] + LARGEUR_ZONE_ODORAT, false);
      }
   }
   void Agir ()
   {
      //
      // Prparation des tats
      //
      descr_.position_ = position ();
      RemplirZoneOdorat (descr_);
      descr_.mode_deplacement_ = mode_;
      DescripteurMonstre apres = descr_;
      MemoireMonstre mem;
      mem.position_ = descr_.position_;
      mem.direction_regard_ = descr_.direction_regard_;
      memoire_.push_back (mem);
      if (memoire_.size () > TAILLE_MEMOIRE_MONSTRE)
         memoire_.pop_front ();
      MemoireMonstre temp[TAILLE_MEMOIRE_MONSTRE];
      copy (memoire_.begin (), memoire_.end (), temp);
      stringstream sstr;
      //
      // Dlgation  la partie tudiante
      //
      ControlerMonstre (sstr, nom (), descr_, apres,
                        temp, static_cast<int>(memoire_.size()));
      //
      // Validation des dcisions
      //
      copy (apres.etats_, apres.etats_ + NB_ETATS_AVATAR, descr_.etats_);
      if (descr_.position_ != apres.position_ && !deplacer (apres.position_))
         sstr << "Mouvement " << descr_.position_
              << " --> " << apres.position_
              << " refus pour le monstre " << nom ();
      else
      {
         descr_.position_ = apres.position_;
         descr_.direction_regard_ = apres.direction_regard_;
      }
      EmettreMessage (sstr.str ());
   }
};

Monstre::Impl *Monstre::Creer
   (const string &nom, const point_t &pos, const mode_deplacement_t mode, Monstre *pAbonne)
{
   Monstre::Impl *p = new Monstre::Impl (nom, pos, mode);
   p->Abonner (pAbonne);
   p->Demarrer ();
   return p;
}

void Monstre::deplacerImpl (const point_t &pos)
{
   pimpl_->deplacerImpl (pos);
   positionner (pos);
}

void Monstre::Liberer (const Monstre::Impl *p) throw()
   { delete p; }

Monstre::~Monstre () throw()
   { pimpl_->Desabonner (static_cast<AbonneDeplacable *>(this)); }

const string & Monstre::nom () const throw ()
   { return pimpl_->nom (); }

template <>
   Deplacable * Monstre::to_<Deplacable>() const throw ()
      { return pimpl_.GetPartageable().get(); }
template <>
   Positionnable* Monstre::to_<Positionnable>() const throw ()
      { return pimpl_.GetPartageable().get(); }