#ifndef PARTIE_ETUDIANTE_H
#define PARTIE_ETUDIANTE_H

#include "PositionTraits.h"
#include "Direction.h"
#include "ServicesSymboles.h"
#include <iosfwd>
using std::ostream;
#include <string>
using std::string;

//
// Utilisez le type point_t pour reprsenter un
// point { X, Y }.
//
// Ex.:
// point_t UnPoint;
// UnPoint.X = 3;
// UnPoint.Y = 3;
//
typedef position_traits::point_t point_t;
//
// Utilisez le type coord_t pour reprsenter
// une coordonne dans un point. Les attributs
// X et Y de point_t sont tous deux de type
// coord_t.
//
typedef position_traits::coord_t coord_t;

//
// Un ZoneRegard reprsente ce que l'avatar voit devant lui
// et autour de lui. Si un avatar est reprsent par le
// symbole A, alors sa zone de regard ira comme suit:
//
// +---+---+---+---+---+
// |   |   |   |   |   |
// +---+---+---+---+---+
// |   |   |   |   |   |
// +---+---+---+---+---+
// |   |   | A |   |   |
// +---+---+---+---+---+
// |   |   |   |   |   |
// +---+---+---+---+---+
// |   |   |   |   |   |
// +---+---+---+---+---+
//
// La zone de regard contient des symboles, que vous 
// pouvez analyser  l'aide des prdicats exposs dans
// le fichier "ServicesSymboles.h" inclus plus haut.
//
const int LARGEUR_ZONE_REGARD = 5,
          HAUTEUR_ZONE_REGARD = 5;
struct ZoneRegard
{
   symbole_t zone_[HAUTEUR_ZONE_REGARD][LARGEUR_ZONE_REGARD];
};

//
// La boussole magique dcrit la distance entre un avatar
// et le plus proche trsor. Ses attributs sont:
// - distance_, qui indique la distance entre l'avatar et
//   le plus proche trsor. La distance est exprime en
//   nombre minimum de pas  faire si on ne trouvait aucun
//   obstacle entre l'avatar et le trsor; et
// - orientation_, qui suggre la direction en vertu de
//   laquelle le nombre de pas requis pour atteindre le
//   plus proche trsor est le moindre.
//
// Note: la boussole magique ne tient pas compte des murs
//
struct BoussoleMagique
{
   int distance_;
   Direction orientation_;
};

//
// Un DescripteurAvatar dcrit l'tat actuel d'un avatar.
// Ses attributs sont:
//
// - position_ (un point_t), reprsentant sa position sur
//   la carte;
// - direction_regard_ (une Direction), indiquant la direction
//   de son regard;
// - observation_, sa zone de regard (voir plus haut);
// - 
// - boussole_magique_, qui indique la distance entre l'avatar
//   et le plus proche trsor. La distance est exprime en
//   nombre minimum de pas  faire si on ne trouvait aucun
//   obstacle entre l'avatar et le trsor;
// - nb_hurlements_proches_, qui indique le nombre de hurlements
//   de monstres  proximit (ici, par  proximit, on entend
//   dans un rayon de RAYON_OUIE_AVATAR cases); et
// - etats_, un tableau d'entiers que vous pourrez utiliser comme
//   bon vous semble.
//
const int NB_ETATS_AVATAR = 10;
const int RAYON_OUIE_AVATAR = 5;
struct DescripteurAvatar
{
   point_t position_;
   Direction direction_regard_;
   ZoneRegard observation_;
   BoussoleMagique boussole_magique_;
   int nb_hurlements_proches_;
   int etats_[NB_ETATS_AVATAR];
};

//
// Tout avatar aura une mmoire parfaite de ce qu'il a
// fait. Chaque moment de sa mmoire sera une instance
// de MemoireAvatar et indiquera sa position, la
// direction de son regard et la valeur indique par sa
// boussole magique.
//
struct MemoireAvatar
{
   point_t position_;
   Direction direction_regard_;
   BoussoleMagique boussole_magique_;
};

//
// Un ZoneOdorat reprsente la zone  l'intrieur de
// laquelle un monstre peut sentir des proies. Si un
// monstre est reprsent par le symbole M, alors sa
// zone d'odorat ira comme suit:
//
// +---+---+---+---+---+---+---+
// |   |   |   |   |   |   |   |
// +---+---+---+---+---+---+---+
// |   |   |   |   |   |   |   |
// +---+---+---+---+---+---+---+
// |   |   |   |   |   |   |   |
// +---+---+---+---+---+---+---+
// |   |   |   | M |   |   |   |
// +---+---+---+---+---+---+---+
// |   |   |   |   |   |   |   |
// +---+---+---+---+---+---+---+
// |   |   |   |   |   |   |   |
// +---+---+---+---+---+---+---+
// |   |   |   |   |   |   |   |
// +---+---+---+---+---+---+---+
//
// Vous remarquerez que l'odorat du monstre peut dtecter
// des proies tout autour de lui, mais que le monstre n'aura
// pas de dtails sur la nature de la proie (la zone est
// faite de boolens, indiquant s'il y a quelque chose ou non,
// et non pas ce qu'est la chose en question.
//
const int LARGEUR_ZONE_ODORAT = 7,
          HAUTEUR_ZONE_ODORAT = 7;
struct ZoneOdorat
{
   bool zone_[HAUTEUR_ZONE_ODORAT][LARGEUR_ZONE_ODORAT];
};

//
// Un mode_deplacement_t reprsentera le mode de dplacement
// d'un monstre donn. Rfrez-vous aux consignes du travail
// pratique pour plus de dtails.
//
typedef int mode_deplacement_t;

//
// Un DescripteurMonstre dcrit l'tat actuel d'un monstre.
// Ses attributs sont:
//
// - position_ (un point_t), reprsentant sa position sur
//   la carte;
// - direction_regard_ (une Direction), indiquant la direction
//   de son regard;
// - chair_fraiche_, sa zone d'odorat pour dtecter les avatars;
// - chair_moite_, sa zone d'odorat pour dtecter les monstres;
// - mode_deplacement_, son mode de dplacement; et
// - etats_, un tableau d'entiers que vous pourrez utiliser comme
//   bon vous semble.
//
const int NB_ETATS_MONSTRE = 10;
struct DescripteurMonstre
{
   point_t position_;
   Direction direction_regard_;
   ZoneOdorat chair_fraiche_;
   ZoneOdorat chair_moite_;
   mode_deplacement_t mode_deplacement_;
   int etats_[NB_ETATS_MONSTRE];
};

//
// Tout monstre aura une mmoire un peu limite de ce qu'il
// a fait (au mieux les TAILLE_MEMOIRE_MONSTRE plus rcents
// moments). Chaque moment de sa mmoire sera une instance
// de MemoireMonstre et indiquera sa position et la direction
// de son regard.
//
const int TAILLE_MEMOIRE_MONSTRE = 10;
struct MemoireMonstre
{
   point_t position_;
   Direction direction_regard_;
};

//
// Fonction ControlerAvatar()
// - Intrants:
//    - nom, le nom de l'avatar (une string);
//    - avant, l'tat de l'avatar au moment o la fonction aura
//      t appele (l'tat avant que votre code n'ait pris une
//      dcision);
//    - memoire, un tableau de MemoireAvatar indiquant la totalit
//      des moments de jeu pour lui jusqu'au moment de l'appel de
//      la fonction; et
//    - nbElemsMemoire, la taille (exprime en nombre d'lments)
//      du tableau Memoire ci-dessus.
// - Extrants:
//    - flux, un flux (que vous pouvez utilisez exactement comme vous
//      utilisez cout) sur lequel vous pourrez crire des messages.
//      Ces messages apparatront  l'cran dans une zone de messages
//      et seront crits sur disque dans un journal (fichier nomm
//      Journal.txt), ce qui vous permettra de dverminer votre oeuvre;
//      de mme que
//    - apres, l'tat de l'avatar tel que vous souhaitez qu'il soit suite
//       l'appel de la fonction (l'tat aprs que votre code ait pris
//      une dcision). Vous pouvez indiquer la direction souhaite pour
//      son regard et la position que vous souhaitez qu'il occupe. Ces
//      valeurs seront acceptes si elles ne sont pas illgales.
// - Rle:
//    - permet de contrler la destine d'un avatar.
//
// Note: si vous dbordez des tableaux, a va planter mchamment et ce
//       sera votre faute. Na na!
//
// Note: si vous faites une boucle infinie dans cette fonction, les autres
//       cratures du jeu continueront quant  elle de fonctionner, alors
//       soyez prudent(e)s!
//
// Note: cette fonction sera appele rgulirement pour permettre  un
//       avatar de prendre une dcision base sur ce qu'il sait, ce qu'il
//       voit et ce qu'il veut faire.  chaque appel, vous recevrez, par
//       le paramtre avant, l'tat o vous en tes prsentement, et
//       vous indiquerez par le paramtre apres l o vous souhaitez en
//       tre lors du prochain appel.
//
// Note: les valeurs que vous indiquerez dans le tableau Etats de Apres
//       seront copies dans le tableau Etats de Avant lors du prochain
//       appel, ce qui vous donnera une certaine latitude pour rflchir.
//
// Note: vous ne pouvez vous dplacer que d'une position  la fois, en X
//       ou en Y (mais pas les deux  la fois). Vous ne pouvez pas
//       traverser les murs. Il n'y a pas de contraintes sur les rotations
//       de la direction du regard. Si vous tentez de tricher, votre
//       mouvement sera refus et vous demeurerez sur place.
//
void ControlerAvatar
   (ostream &flux, const string &nom,
                   const DescripteurAvatar &avant,
                   DescripteurAvatar &apres,
                   const MemoireAvatar memoire[],
                   const int nbElemsMemoire);

//
// Fonction ControlerMonstre()
// - Intrants:
//    - nom, le nom du monstre (une string);
//    - avant, l'tat du monstre au moment o la fonction aura t
//      appele (l'tat avant que votre code n'ait pris une dcision);
//    - memoire, un tableau de MemoireMonstre indiquant les plus rcents
//      moments de jeu pour lui jusqu'au moment de l'appel de la fonction; et
//    - nbElemsMemoire, la taille (exprime en nombre d'lments) du tableau
//      Memoire ci-dessus (PRUDENCE: cette taille peut tre plus petite que
//      TAILLE_MEMOIRE_MONSTRE, surtout au dbut du jeu!).
// - Extrants:
//    - flux, un flux (que vous pouvez utilisez exactement comme vous
//      utilisez cout) sur lequel vous pourrez crire des messages.
//      Ces messages apparatront  l'cran dans une zone de messages
//      et seront crits sur disque dans un journal (fichier nomm
//      Journal.txt), ce qui vous permettra de dverminer votre oeuvre;
//      de mme que
//    - apres, l'tat du monstre tel que vous souhaitez qu'il soit suite
//       l'appel de la fonction (l'tat aprs que votre code ait pris
//      une dcision). Vous pouvez indiquer la direction souhaite pour
//      son regard et la position que vous souhaitez qu'il occupe. Ces
//      valeurs seront acceptes si elles ne sont pas illgales.
// - Rle:
//    - permet de contrler la destine d'un monstre. Sera appele pour
//      chaque monstre avec des tats avant/ apres distincts.
//
// Note: si vous dbordez des tableaux, a va planter mchamment et ce
//       sera votre faute. Na na!
//
// Note: si vous faites une boucle infinie dans cette fonction, les autres
//       cratures du jeu continueront quant  elle de fonctionner, alors
//       soyez prudent(e)s!
//
// Note: cette fonction sera appele rgulirement pour permettre  un
//       monstre de prendre une dcision base sur ce qu'il sait, ce qu'il
//       sent et ce qu'il veut faire.  chaque appel, vous recevrez, par
//       le paramtre avant, l'tat o vous en tes prsentement, et
//       vous indiquerez par le paramtre apres l o vous souhaitez en
//       tre lors du prochain appel.
//
// Note: les valeurs que vous indiquerez dans le tableau Etats de Apres
//       seront copies dans le tableau Etats de Avant lors du prochain
//       appel, ce qui vous donnera une certaine latitude pour rflchir.
//
// Note: vous ne pouvez vous dplacer que d'une position  la fois, en X
//       ou en Y (mais pas les deux  la fois). Vous ne pouvez pas
//       traverser les murs. Il n'y a pas de contraintes sur les rotations
//       de la direction du regard. Si vous tentez de tricher, votre
//       mouvement sera refus et vous demeurerez sur place.
//
void ControlerMonstre
   (ostream &flux, const string &nom,
                   const DescripteurMonstre & avant,
                   DescripteurMonstre &apres,
                   const MemoireMonstre memoire[],
                   const int nbElemsMemoire);

//
// Fonction Hurler()
// - Intrants:
//    - nom, le nom du monstre hurlant;
//    - desc, le descripteur du monstre hurlant.
// -Extrants: --
// - Rle:
//    - faire hurler un monstre. Utile lorsqu'un monstre se trouve 
//      proximit d'un avatar.
//
void Hurler (const string &nom, const DescripteurMonstre &desc);

#endif