#ifndef PTR_PARTAGE_H
#define PTR_PARTAGE_H

#include "Testable.h"
#include <utility> // std::swap()

struct Detruire
{
   template <class T>
      void operator () (const T p) const
         { delete p; }
   template <class T>
      void operator () (const volatile T p) const volatile
         { delete p; }
};
struct DetruireTableau
{
   template <class T>
      void operator () (const T p) const
         { delete [] p; }
   template <class T>
      void operator () (const volatile T p) const volatile
         { delete [] p; }
};

template <class T, class PolitiqueSuppression = Detruire>
   class Partageable
   {
   public:
      typedef unsigned long cpt_type;
      typedef cpt_type * ptr_cpt_type;
   private:
      mutable cpt_type * pcptref_;
   public:
      typedef typename T value_type;
      typedef typename value_type * pointer;
   private:
      pointer ptr_;
      PolitiqueSuppression polsupp_;
   public:
      Partageable (pointer p = 0, PolitiqueSuppression ps = PolitiqueSuppression ())
         : ptr_ (p), pcptref_ (0), polsupp_ (ps)
      {
         if (ptr_) Create ();
      }
      Partageable(const Partageable & p, PolitiqueSuppression ps = PolitiqueSuppression ()) throw ()
         : ptr_ (p.get ()), pcptref_ (0), polsupp_ (ps)
      {
         p.Remap (pcptref_);
      }
      void Remap (ptr_cpt_type &p) const  throw ()
      {
         if (p = pcptref_) AddRef ();
      }
      template <class U>
         Partageable(const Partageable<U> & p, PolitiqueSuppression ps = PolitiqueSuppression ()) throw ()
            : ptr_ (0), pcptref_ (0), polsupp_ (ps)
         {
            if (ptr_ = dynamic_cast<T*> (p.get ()))
               p.Remap (pcptref_);
         }
      void swap (Partageable &p) throw ()
      {
         std::swap (ptr_, p.ptr_);
         std::swap (pcptref_, p.pcptref_);
      }
      const Partageable & operator= (const Partageable &p) throw ()
      {
         Partageable (p).swap (*this);
         return *this;
      }
      template <class U>
         const Partageable & operator= (const Partageable<U> &p) throw ()
         {
            Partageable (p).swap (*this);
            return *this;
         }
      ~Partageable () throw ()
         { if (pcptref_) Release (); }
   private:
      void Create () const
         { pcptref_ = new cpt_type (1); }
      void AddRef () const throw()
         { ++(*pcptref_); }
      void Release () const throw ()
      {
         if (--(*pcptref_) == 0)
         {
            delete pcptref_;
            polsupp_ (ptr_);
         }
      }
   public:
      // ICI: patch
      pointer get () const throw ()
         { return ptr_; }
   };

namespace std
{
   template <class T>
      void swap (Partageable <T> &a, Partageable <T> &b) throw ()
         { a.swap (b); }
}

template <class T, class PolitiqueSuppression = Detruire>
   class PtrPartage
      : public Testable
   {
      Partageable <T, PolitiqueSuppression> ptr_;
   public:
      const Partageable<T, PolitiqueSuppression> &
         GetPartageable () const throw()
            { return ptr_; }
   private:
      class AutoRelacheur
      {
         Partageable <T, PolitiqueSuppression> temp_;
      public:
         AutoRelacheur (const Partageable <T, PolitiqueSuppression> & p) throw()
            : temp_ (p)
         {
         }
         ~AutoRelacheur () throw()
            { }
      };

      T *get () const throw()
         { return ptr_.get (); }
   public:
      PtrPartage (T *p = 0)
         : ptr_ (p)
      {
      }
      PtrPartage (const PtrPartage &p) throw()
         : ptr_ (p.GetPartageable ())
      {
      }
      template <class U, class Pol>
         PtrPartage (const PtrPartage <U, Pol> &p) throw()
            : ptr_ (p.GetPartageable ())
         {
         }
      void swap (PtrPartage &p) throw ()
         { std::swap (ptr_, p.ptr_); }
      const PtrPartage & operator= (const PtrPartage &p) throw ()
      {
         PtrPartage (p).swap (*this);
         return *this;
      }
      template <class U, class Pol>
         const PtrPartage & operator= (const PtrPartage<U, Pol> &p) throw ()
         {
            PtrPartage (p).swap (*this);
            return *this;
         }
      ~PtrPartage () throw()
         { }
      T & operator * ()
         { return *(get ()); }
      const T & operator * () const
         { return *(get ()); }
      T * operator -> ()
      {
         AutoRelacheur ar (ptr_);
         return get();
      }
      const T * operator -> () const
      {
         AutoRelacheur ar (ptr_);
         return get();
      }
   protected:
      bool_type testValidite () const
         { return get ()? true_value() : false_value(); }
   };

template <class T, class U, class Pol>
   PtrPartage <T, Pol> partageable_cast (const PtrPartage<U, Pol> &p)
      { return PtrPartage <T, Pol> (p); }

namespace std
{
   template <class T, class Pol>
      void swap (PtrPartage <T, Pol> &a, PtrPartage <T, Pol> &b) throw()
         { a.swap (b); }
}

#endif

