#ifndef SERVICE_ABONNEMENT_H
#define SERVICE_ABONNEMENT_H

#include "Mutex.h"
#include <vector>
using std::vector;
#include <memory>
using std::allocator;
#include <cassert>

template <class T, template <class, class> class C = vector, class A = allocator<T> >
   class ServiceAbonnement
   {
      typedef C<T*,A> conteneur_t;
      conteneur_t abonnes_;
      Mutex mutex_;
   public:
      typedef typename
         conteneur_t::value_type value_type;
   protected:
      void AbonnerImpl (value_type p)
      {
         assert (p); // parce que protected, alors mieux vaut tre prudents
         abonnes_.push_back (p);
      }
      void DesabonnerImpl (value_type p)
      {
         assert (p); // parce que protected, alors mieux vaut tre prudents
         abonnes_.erase (remove (begin (), end (), p));
      }
   public:
      typedef typename
         conteneur_t::iterator iterator;
      typedef typename
         conteneur_t::size_type size_type;
      typedef typename
         conteneur_t::const_iterator const_iterator;
      iterator begin () throw ()
         { return abonnes_.begin(); }
      iterator end () throw ()
         { return abonnes_.end (); }
      const_iterator begin () const throw ()
         { return abonnes_.begin(); }
      const_iterator end () const throw ()
         { return abonnes_.end (); }
      size_type size () const throw ()
         { return abonnes_.size (); }
      void Abonner (value_type p) volatile
      {
         Autoverrou av(mutex_);
         const_cast<ServiceAbonnement *>(this)->AbonnerImpl (p);
      }
      void Desabonner (value_type p) volatile
      {
         Autoverrou av(mutex_);
         const_cast<ServiceAbonnement *>(this)->DesabonnerImpl (p);
      }
   };

#endif