#ifndef TEMPS_H
#define TEMPS_H

//
// Rfrentiels
//
class Seconde {};
class Milliseconde {}; // rfrentiel neutre choisi
class Microseconde {};

//
// Traits
//

#include "locale_helper.h"
#include <string>
using std::basic_string;

struct base_traits_temps
{
   typedef double value_type;
   static value_type default_value ()
      { return value_type (); }
protected:
   template <class CType>
      static basic_string<CType> nom (const basic_string<char> &s)
         { return convertir<CType>(s); }
};

template <class>
   struct traits_temps;
template <>
   struct traits_temps<Seconde>
      : base_traits_temps
   {
      template <class CType>
         static basic_string<CType> nom ()
            { return base_traits_temps::nom<CType>("Seconde"); }
      static value_type to_neutral (const value_type val)
         { return val / static_cast<value_type>(1000); }
      static value_type from_neutral (const value_type val)
         { return val * static_cast<value_type>(1000); }
   };
template <>
   struct traits_temps<Milliseconde>
      : base_traits_temps
   {
      template <class CType>
         static basic_string<CType> nom ()
            { return base_traits_temps::nom<CType>("Milliseconde"); }
      static value_type to_neutral (const value_type val)
         { return val; }
      static value_type from_neutral (const value_type val)
         { return val; }
   };
template <>
   struct traits_temps<Microseconde>
      : base_traits_temps
   {
      template <class CType>
         static basic_string<CType> nom ()
            { return base_traits_temps::nom<CType>("Microseconde"); }
      static value_type to_neutral (const value_type val)
         { return val * static_cast<value_type>(1000); }
      static value_type from_neutral (const value_type val)
         { return val / static_cast<value_type>(1000); }
   };

//
// Temps
//
#include <utility> // std::swap

template <class D, class S>
   typename traits_temps<D>::value_type
      temps_cast (const typename traits_temps<S>::value_type val)
   {
      return traits_temps<D>::from_neutral (traits_temps<S>::to_neutral(val));
   }

#include "relation.h"

template <class R>
   class Temps
      : relation::equivalence<Temps<R> >,
        relation::ordre<Temps<R> >
   {
   public:
      typedef R referentiel;
      typedef typename
         traits_temps<R>::value_type value_type;
   private:
      value_type valeur_;
   public:
      Temps (const value_type val = traits_temps<R>::default_value ())
         : valeur_(val)
      {
      }
      template <class U>
         Temps (const Temps<U> &t)
            : valeur_(temps_cast<R, U>(t.value ()))
         {
         }
      void swap (Temps &t)
         { std::swap (valeur_, t.valeur_); }
      template <class U>
         Temps & operator= (const Temps<U> &t)
         {
            Temps (t).swap (*this);
            return *this;
         }
      value_type value () const
         { return valeur_; }
      bool operator== (const Temps &t) const
         { return value () == t.value (); }
      bool operator< (const Temps &t) const
         { return value () < t.value (); }
      Temps & operator++ ()
      {
         ++valeur_;
         return *this;
      }
      Temps operator++ (int)
      {
         Temps temp (*this)
         operator++();
         return temp;
      }
   };

#include <iosfwd>
using std::basic_ostream;

template <class CType, class R>
   basic_ostream<CType> operator<<
      (basic_ostream<CType> &os, const Temps<R> &t)
   {
      return os << t.value () << char_convert<CType>(' ')
                << traits_temps<R>::nom<CType>();
   }

#endif

